// SPDX-License-Identifier: GPL-3.0-or-later

package tengine

import (
	"net/http"
	"net/http/httptest"
	"os"
	"testing"

	"github.com/netdata/netdata/go/plugins/plugin/go.d/agent/module"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

var (
	dataConfigJSON, _ = os.ReadFile("testdata/config.json")
	dataConfigYAML, _ = os.ReadFile("testdata/config.yaml")

	dataStatusMetrics, _ = os.ReadFile("testdata/status.txt")
)

func Test_testDataIsValid(t *testing.T) {
	for name, data := range map[string][]byte{
		"dataConfigJSON":    dataConfigJSON,
		"dataConfigYAML":    dataConfigYAML,
		"dataStatusMetrics": dataStatusMetrics,
	} {
		require.NotNil(t, data, name)
	}
}

func TestCollector_ConfigurationSerialize(t *testing.T) {
	module.TestConfigurationSerialize(t, &Collector{}, dataConfigJSON, dataConfigYAML)
}

func TestCollector_Cleanup(t *testing.T) {
	New().Cleanup()
}

func TestCollector_Init(t *testing.T) {
	collr := New()

	require.NoError(t, collr.Init())
}

func TestCollector_Check(t *testing.T) {
	ts := httptest.NewServer(
		http.HandlerFunc(
			func(w http.ResponseWriter, r *http.Request) {
				_, _ = w.Write(dataStatusMetrics)
			}))
	defer ts.Close()

	collr := New()
	collr.URL = ts.URL
	require.NoError(t, collr.Init())
	assert.NoError(t, collr.Check())
}

func TestCollector_CheckNG(t *testing.T) {
	collr := New()

	collr.URL = "http://127.0.0.1:38001/us"
	require.NoError(t, collr.Init())
	assert.Error(t, collr.Check())
}

func TestCollector_Charts(t *testing.T) { assert.NotNil(t, New().Charts()) }

func TestCollector_Collect(t *testing.T) {
	ts := httptest.NewServer(
		http.HandlerFunc(
			func(w http.ResponseWriter, r *http.Request) {
				_, _ = w.Write(dataStatusMetrics)
			}))
	defer ts.Close()

	collr := New()
	collr.URL = ts.URL
	require.NoError(t, collr.Init())
	require.NoError(t, collr.Check())

	expected := map[string]int64{
		"bytes_in":                 5944,
		"bytes_out":                20483,
		"conn_total":               354,
		"http_200":                 1536,
		"http_206":                 0,
		"http_2xx":                 1536,
		"http_302":                 43,
		"http_304":                 0,
		"http_3xx":                 50,
		"http_403":                 1,
		"http_404":                 75,
		"http_416":                 0,
		"http_499":                 0,
		"http_4xx":                 80,
		"http_500":                 0,
		"http_502":                 1,
		"http_503":                 0,
		"http_504":                 0,
		"http_508":                 0,
		"http_5xx":                 1,
		"http_other_detail_status": 11,
		"http_other_status":        0,
		"http_ups_4xx":             26,
		"http_ups_5xx":             1,
		"req_total":                1672,
		"rt":                       1339,
		"ups_req":                  268,
		"ups_rt":                   644,
		"ups_tries":                268,
	}

	assert.Equal(t, expected, collr.Collect())
}

func TestCollector_InvalidData(t *testing.T) {
	ts := httptest.NewServer(
		http.HandlerFunc(
			func(w http.ResponseWriter, r *http.Request) {
				_, _ = w.Write([]byte("hello and goodbye"))
			}))
	defer ts.Close()

	collr := New()
	collr.URL = ts.URL
	require.NoError(t, collr.Init())
	assert.Error(t, collr.Check())
}

func TestCollector_404(t *testing.T) {
	ts := httptest.NewServer(
		http.HandlerFunc(
			func(w http.ResponseWriter, r *http.Request) {
				w.WriteHeader(http.StatusNotFound)
			}))
	defer ts.Close()

	collr := New()
	collr.URL = ts.URL
	require.NoError(t, collr.Init())
	assert.Error(t, collr.Check())
}
